use serenity::all::{CommandDataOption, CommandDataOptionValue, CommandInteraction, CommandOptionType, CreateCommandOption};
use serenity::{builder::CreateCommand, client::Context};
use skynet_discord_bot::common::database::{get_server_config, DataBase, Servers};
use skynet_discord_bot::common::set_roles::normal::update_server;
use skynet_discord_bot::common::wolves::cns::get_wolves;
use sqlx::{Error, Pool, Sqlite};

pub async fn run(command: &CommandInteraction, ctx: &Context) -> String {
  let wolves_api = if let Some(CommandDataOption {
    value: CommandDataOptionValue::String(key),
    ..
  }) = command.data.options.first()
  {
    key.to_string()
  } else {
    return "Please provide a wolves API key".to_string();
  };

  let role_current = if let Some(CommandDataOption {
    value: CommandDataOptionValue::Role(role),
    ..
  }) = command.data.options.get(1)
  {
    role.to_owned()
  } else {
    return "Please provide a valid role for ``Role Current``".to_string();
  };

  let role_past = if let Some(CommandDataOption {
    value: CommandDataOptionValue::Role(role),
    ..
  }) = command.data.options.get(5)
  {
    Some(role.to_owned())
  } else {
    None
  };

  let bot_channel_id = if let Some(CommandDataOption {
    value: CommandDataOptionValue::Channel(channel),
    ..
  }) = command.data.options.get(2)
  {
    channel.to_owned()
  } else {
    return "Please provide a valid channel for ``Bot Channel``".to_string();
  };

  let db_lock = {
    let data_read = ctx.data.read().await;
    data_read.get::<DataBase>().expect("Expected Databse in TypeMap.").clone()
  };
  let db = db_lock.read().await;

  let server_data = Servers {
    server: command.guild_id.unwrap_or_default(),
    wolves_api,
    wolves_id: 0,
    role_past,
    role_current,
    member_past: 0,
    member_current: 0,
    bot_channel_id,
  };

  match add_server(&db, ctx, &server_data).await {
    Ok(_) => {}
    Err(e) => {
      println!("{:?}", e);
      return format!("Failure to insert into Servers {:?}", server_data);
    }
  }

  "Added/Updated server info".to_string()
}

pub fn register() -> CreateCommand {
  CreateCommand::new("add")
    .description("Enable the bot for this discord")
    .default_member_permissions(serenity::model::Permissions::MANAGE_GUILD)
    .add_option(CreateCommandOption::new(CommandOptionType::String, "api_key", "UL Wolves API Key").required(true))
    .add_option(CreateCommandOption::new(CommandOptionType::Role, "role_current", "Role for Current members").required(true))
    .add_option(CreateCommandOption::new(CommandOptionType::Channel, "bot_channel", "Safe space for folks to use the bot commands.").required(true))
    .add_option(CreateCommandOption::new(CommandOptionType::Role, "role_past", "Role for Past members").required(false))
}

async fn add_server(db: &Pool<Sqlite>, ctx: &Context, server: &Servers) -> Result<Option<Servers>, Error> {
  let existing = get_server_config(db, &server.server).await;
  let role_past = server.role_past.map(|x| x.get() as i64);

  let insert = sqlx::query_as::<_, Servers>(
    "
      INSERT OR REPLACE INTO servers (server, wolves_api, role_past, role_current, bot_channel_id)
      VALUES (?1, ?2, ?3, ?4, ?5)
    ",
  )
  .bind(server.server.get() as i64)
  .bind(&server.wolves_api)
  .bind(role_past)
  .bind(server.role_current.get() as i64)
  .bind(server.bot_channel_id.get() as i64)
  .fetch_optional(db)
  .await;

  // if the entry does not exist already tehn do a user update
  let (update, current_remove, current_role, past_remove, past_role) = match &existing {
    None => (true, false, None, false, None),
    Some(x) => {
      let mut result = (false, false, None, false, None);
      if x.wolves_api != server.wolves_api {
        result.0 = true;
      }
      if x.role_current != server.role_current {
        result.0 = true;
        result.1 = true;
        result.2 = Some(x.role_current);
      }
      if x.role_past != server.role_past {
        result.0 = true;
        result.3 = true;
        result.4 = x.role_past;
      }
      result
    }
  };

  // update all users
  if update {
    // handle wolves api here
    get_wolves(ctx).await;

    let mut roles_remove = vec![];
    if current_remove {
      roles_remove.push(current_role)
    }
    if past_remove {
      roles_remove.push(past_role)
    }
    update_server(ctx, server, &roles_remove, &[]).await;
  }

  insert
}
